// -*- mode: c++ -*-
//  copyright (C) 2001 Makoto Onizuka, NTT CyberSpace Labolatories
//  $Id: XPath.cxx,v 1.3 2002/08/13 12:59:46 monizuka Exp $

#ifndef WIN32
#include <iostream.h>
#else
#include <iostream>
#endif

#include "Error.h"
#include "Query.h"
#include "Variable.h"
#include "XPath.h"
#include "List.cxx"

template class List<LocationStep>;
template class List<XPath>;
template class List<Predicate>;

static PString	memoryPool[PSTRING_STATIC_MEMORY];
static unsigned int	memoryCount = 0;

void *
PString::operator new(size_t s){
  return &memoryPool[memoryCount++];
}

void
PString::operator delete(void * p){
  memoryCount--;
}

const char * 
PString::getString(void) {
  switch(type){
  case T_XMLSTRING:
#if defined(XERCES_PARSER)
    if (str != 0) return str;
    str = XMLString::transcode(xstr);
    return str;
#endif
  case T_STRING:{
    char * s = (char *)malloc(length+1);
    strncpy(s, str, length);
    s[length] = '\0';
    return (const char *)s;
  }
#if defined(TOKEN_PARSER)
  case T_TOKEN:
    return g_ptt->StrFromXTOKEN(token);
#endif
  default:
    cerr << "inernal error@PString::getString()" << endl;
    return 0;
  }
}

ostream& operator<<(ostream& target, PString * toDump)
{
    target << toDump->getString();
    return target;
}

// Error messages for XPath class.
char * XPath::errmes[] = {
  "",
  "Memory allocation error",
  "Other class error",
  "System internal error"
};

XPath::XPath():Base(),var(0),automata(0),lsteps(new List<LocationStep>),parent(0) {}

void XPath::deleteLast(void)
{
    lsteps->deleteItem(lsteps->getCount()-1);
    LocationStep * sx = lsteps->last();
    sx->next = 0;
}

LocationStep * XPath::pop(void)
{
    return lsteps->pop();
}

void XPath::print(ofstream * ofs)
{
    for (unsigned int i = 0; i < lsteps->getCount(); i++)
    {
        LocationStep * sp = lsteps->getItem(i);
        switch(sp->getType())
        {
      case LocationStep::DoubleSlash:
	* ofs << "/";
	break;
      case LocationStep::Period:
	* ofs << ".";
	break;
      case LocationStep::DoublePeriod:
	* ofs << "..";
	break;
      default:			// Element or Attribute
	if (sp->path) * ofs << sp->path;
	else {
	  * ofs << "(";
	  for (unsigned int j = 0; j < sp->altPath->getCount(); j++){
	    XPath * p = sp->altPath->getItem(j);
	    p->print(ofs);
	    if (j + 1 < sp->altPath->getCount()) * ofs << "|";
	  }
	  * ofs << ")";
	}
      }
      if (i + 1 < lsteps->getCount()) * ofs << '/';
    }
}

//-------------------------------------------------------------------------------------------------
// LocationStep
//-------------------------------------------------------------------------------------------------

// Error messages for LocationStep class.
char * LocationStep::errmes[] = {
  "",
  "Memory allocation error",
  "Other class error",
  "Multiple index predicate \"element[pred]...[n]...[m]...[pred]\" is not proceeded.\n Please use single index predicate.",
  "...//element[index] is not supported yet",
  "Not supported",
  "System internal error"
};


LocationStep::LocationStep(XPath * xp, nodeType n, char * p) : Base(), parent(xp), next(0), nType(n), pType(P_NoPredicate), altPath(0), preds(new List<Predicate>()),index(0)
{
    if (n == Element)
    {
        path = strdup(p);
#if defined(TOKEN_PARSER)
        token = g_ptt->XTOKENFromStr(p, XST_ELEMENT);
#endif
    }
    else if (n == Attribute)
    {
        path = strdup(p);			// p contains '@'
#if defined(TOKEN_PARSER)
        token = g_ptt->XTOKENFromStr(p+1, XST_ATTRIBUTE);
#endif
    }
    else
    {
        path = NULL;
        altPath = new List<XPath>();
    }
        
    LocationStep * last = xp->getLastLStep();
    if (last != NULL)
        last->next = this; // if this is not top
    xp->insertLStep(this);
}

LocationStep::~LocationStep()
{
    if (path != NULL) 
        free(path);
    else 
    {
        altPath->delItems();
        delete altPath;
    }
    
    for (unsigned int i = 0;i<preds->getCount();i++)
        delete preds->getItem(i);
    delete preds;
}

void LocationStep::simplePred2string(std::string & s)
{
    for (unsigned int ulPred = 0 ; ulPred < preds->getCount() ; ulPred++)
    {
        Predicate * pPred = preds->getItem(ulPred);
        switch (pPred->getType())
        {
            case Predicate::T_INDEX:
            case Predicate::T_DOUBLESLASH_INDEX:
            case Predicate::T_PREDICATE:
                // do nothing
                break;
            case Predicate::T_SIMPLE_PREDICATE:
                s.append("[");
                pPred->pred2string(s);
                s.append("]");
                break;
        }
    }
}

void LocationStep::indexPred2string(std::string & s)
{
    for (unsigned int ulPred = 0 ; ulPred < preds->getCount() ; ulPred++)
    {
        Predicate * pPred = preds->getItem(ulPred);
        switch (pPred->getType())
        {
            case Predicate::T_INDEX:
            case Predicate::T_DOUBLESLASH_INDEX:
                if (ulPred==0) 
                {
                    s.append("[");
                    pPred->pred2string(s);
                    s.append("]");
                }
                else 
                    throw CXmlToolkitException(ERROR_NOT_IMPLEMENTED);
                break;
            case Predicate::T_PREDICATE:
            case Predicate::T_SIMPLE_PREDICATE:
				// do nothing
                break;
        }
    }
}

bool LocationStep::match(PString * ps)
{
    if (nType == AnyElement) return true;
    if (ps == 0 || path == 0) return false;
#if defined(TOKEN_PARSER)
    if (ps->comp(token)==0) return true;
#else	
    if (ps->comp(path)==0) return true;
#endif
    else return false;
}

void LocationStep::setType(void)
{
    for (unsigned int i=0;i<preds->getCount();i++){
      Predicate * p = preds->getItem(i);
      switch(p->getType()){
      case Predicate::T_INDEX:
	if (index)
        throw CXmlToolkitException(errmes[E_MULTIPLE_INDEXPREDICATE]);

    pType = P_Index;
	index = p;		// cache the predicate as "index"
	break;
      case Predicate::T_SIMPLE_PREDICATE:
	if (pType==P_NoPredicate) pType = P_SimplePredicate;
	break;
      case Predicate::T_PREDICATE:
	if (pType==P_NoPredicate||pType==P_SimplePredicate) pType = P_Predicate;
	break;
      default:		// T_DOUBLESLASH_INDEX
	break;		// do nothing
      } // end switch
    }
    return;
}

void LocationStep::setPredTypeDoubleSlash(void)
{
    if (index) 
    {
        index->setType(Predicate::T_DOUBLESLASH_INDEX);
        throw CXmlToolkitException(errmes[E_DOUBLESLASH_INDEX]);
    }
}

char * LocationStep::getLabel(void)
{
    switch (nType)
    {
    case Element:
    case Attribute:
        return path;
    case Text:
        return "text()";
    case Period:
        return ".";
    case DoublePeriod:
        return "..";
    case AnyElement:
        return "*";
    case AnyAttribute:
        return "@*";
    case DoubleSlash:
        return "//";
    case DocumentRoot:
        return "/";
    default:
        throw CXmlToolkitException(ERROR_UNEXPECTED);
    }
}